# CheapPopup by David Bowland
# ./addons/eventscripts/_libs/python/cheappopup.py

""" Our popups suck less(tm) """

from __future__ import division

import cfglib
import cmdlib
import collections
import es
import gamethread
import langlib
import math

from path import path

import psyco
psyco.full()


"""
* Functions available through CheapPopup:
   NOTE: Each function below is also available through the CheapPopup class

sendMenu(userid, duration, text, callback=None)
   "
   Sends a single menu to a player
   Optional callback will be called with menu selection
   "
   userid = player to send menu
   duration = integer number of seconds to display menu, 0 = until selection
   text = text to display in menu (\n = newline)
   callback = function to call when a selection is made: callback(userid, selection)
      where "userid" is the integer userid of the player making the selection
      and "selection" is the integer selection number (0 through 9 inclusive) or None for no selection (timeout/menu closed)


sendSelectionMenu(userid, items, callback, langobj=SelectionLanguages,
                  firstback=None, lastnext=None, nohighlight=None)
   "
   Sends a series of menus to a player prompting selection
   Returns selection through callback
   "
   userid = player to send series of menus
   items = object containing string items for selection
      when an instance of collections.Sequence (tuple, list, etc) is given, selections are displayed in order and an integer index selection is returned
      when an instance of collections.Set (set, frozenset, etc) is given, selections are displayed sorted and the string selection is returned
      when an instance of collections.Mapping (dictionary, etc) is given, selections are displayed sorted and the value corresponding to the key selected is returned
      if an object passed has the "order" attribute, that sequence will be used to dictate display order
         (see SelectionDict and SelectionSet below)
   callback = function to call when a selection is made: callback(userid, selection)
      where "userid" is the integer userid of the player making the selection
      and "selection" is an object determined by the object passed to "items" (above) or None for no selection (including menu closed)
   langobj = object containing language strings used to build menus
      when a string is given, that string is used for the "header" language item
      (see SelectionLanguages and SelectionLanguagesLanglib)
   firstback = function to call if the first "back" selection is made: firstback(userid)
      where "userid" is the integer userid of the player making the selection
      NOTE: When this argument is omitted this selection is replaced by a selection to go to the last page of items
   lastnext = function to call if the last "next" selection is made: lastnext(userid)
      where "userid" is the integer userid of the player making the selection
      NOTE: When this argument is omitted this selection is replaced by a selection to go to the first page of items
   nohighlight = container object dictating selections that should not be highlighted (appear yellow/orange on the menu)
      when omitted, given None or False all selections appear highlighted
      when given a container, selection items appearing in the container are not highlighted
      when True, no selection appears highlighted


sendESCMenu(userid, duration, title, text)
   " Sends a menu of text to a player, ESC must be pressed to view "
   userid = player to send menu
   duration = integer number of seconds to display menu (10 <= duration <= 200, this limit is Valve-imposed)
   title = title to use for menu (this text also appears to alert the player of the menu)
   text = text to display in menu


sendESCButtonMenu(userid, duration, title, text, items, callback)
   "
   Sends a multi-button GUI menu to the player, ESC must be pressed to view
   Returns selection through callback
   NOTE: This menu's input functionality conflicts with the popup menu's functionality.
      Therefore, if a popup is open when an ESCButtonMenu is sent, the popup will be closed, and vice-versa.
   "
   userid = player to send menu
   duration = integer number of seconds to display menu (10 <= duration <= 200, this limit is Valve-imposed)
   title = title to use for menu (this text also appears to alert the player of the menu)
   text = text to display in menu (prompt for selection)
   items = tuple or list of up to eight strings for button text
   callback = function to call when a selection is made: callback(userid, selection)
      where "userid" is the integer userid of the player making the selection
      and "selection" is the integer button number the player selected (1 through 8 inclusive) or None for no selection (timeout/menu closed)


sendESCButtonSelectionMenu(userid, duration, items, callback, langobj=SelectionLanguages,
                           firstback=None, lastnext=None)
   "
   Sends a series of multi-button GUI menus to the player prompting selection, ESC must be pressed to view
   Returns selection through callback
   NOTE: This function uses ESCButtonMenu and therefore suffers from the same conflicts with popups.
      See the notes for sendESCButtonMenu for further information.
   "
   userid = player to send series of menus
   duration = integer number of seconds to display each menu (10 <= duration <= 200, this limit is Valve-imposed)
   items = object containing string items for selection
      when an instance of collections.Sequence (tuple, list, etc) is given, selections are displayed in order and an integer index selection is returned
      when an instance of collections.Set (set, frozenset, etc) is given, selections are displayed sorted and the string selection is returned
      when an instance of collections.Mapping (dictionary, etc) is given, selections are displayed sorted and the value corresponding to the key selected is returned
      if an object passed has the "order" attribute, that sequence will be used to dictate display order
         (see SelectionDict and SelectionSet below)
   callback = function to call when a selection is made: callback(userid, selection)
      where "userid" is the integer userid of the player making the selection
      and "selection" is an object determined by the object passed to "items" (above) or None for no selection (including timeout or menu closed)
   langobj = object containing language strings used to build menus
      when a string is given, that string is used for the "title" and "text" language items
      (see SelectionLanguages and SelectionLanguagesLanglib)
   firstback = function to call if the first "back" selection is made: firstback(userid)
      where "userid" is the integer userid of the player making the selection
      NOTE: When this argument is omitted this selection is replaced by a selection to go to the last page of items
   lastnext = function to call if the last "next" selection is made: lastnext(userid)
      where "userid" is the integer userid of the player making the selection
      NOTE: When this argument is omitted this selection is replaced by a selection to go to the first page of items


* Classes available through CheapPopup:

CheapPopup()
   "
   Container class for associated menus
   The advantage of this class being the "close" functions only affect menus opened by the same instance
   On player_disconnect any popup delays associated with the disconnecting player are removed
   On es_map_start all popup delays associated with the instance are removed
   "
   sendMenu # function documented above
   sendSelectionMenu # function documented above
   sendESCMenu # function documented above
   sendESCInputMenu # function documented above
   sendESCButtonMenu # function documented above
   sendESCButtonSelectionMenu # function documented above
   sendCustomMenu(Menu, userid, *a, **kw)
      Sends a custom Popup implementation to the given userid
   sendCustomESCMenu(Menu, userid, *a, **kw)
      Sends a custom ESCMenu implementation to the given userid
   closeMenu(userid)
      Closes the non-ESC menu open to the given userid
   closeESCMenu(userid)
      Closes the ESC menu open to the given userid
   closeAll()
      Closes all menus associated with the CheapPopup instances
   clear()
      Removes all popup delays associated with the CheapPopup instance but DOES NOT CLOSE MENUS


Popup()
   "
   Base Popup class
   All send*Menu functions return an instance of this class or a subclass
   "
   close()
      Closes the popup if it is active
   cancelDelay()
      Cancels any delay associated with the popup (useful if the popup is overwritten by another menu system)
   isActive()
      Returns True if the popup is active (displayed) otherwise returns False


SelectionLanguages()
   " Container object for strings associated with a SelectionPopup "
   header = 'Make a selection:'
   noselections = 'No selections'
   separator1 = ' '
   separator2 = ' '
   back = 'Back'
   next = 'Next'
   firstback = 'Last page'
   lastnext = 'First page'
   exit = 'Exit'


ESCSelectionLanguages()
   " Container object for strings associated with an ESCButtonSelectionMenu "
   title = 'Make a selection:'
   text = 'Please make a selection:'
   back = 'Back'
   next = 'Next'
   firstback = 'Last page'
   lastnext = 'First page'


SelectionLanguagesLanglib(langobj=None, lang=None, knownas=None)
   "
   Container object that pulls strings for SelectionPopup or ESCButtonSelectionMenu from a langlib object
   When "knownas" is provided a mapping, string names will first be searched for in "knownas"
      Such that if "knownas" is provided: {'header': 'my header'}
      When "header" is displayed, the langlib item "my header" will be used
   "


SelectionSet(itr=None)
   "
   Subclass of "set" that remembers the order items are added
   When initialized with an iterator, items are added in sorted(...) order
   "


SelectionDict(itr=None)
   "
   Subclass of "dict" that remembers the order keys are added
   When initialized with an iterator, keys are added in sorted(...) order
   "
   
"""

__all__ = [
 # Base popup
 'Popup',
 'sendMenu',
 # Selection popup
 'SelectionLanguages',
 'SelectionLanguagesLanglib',
 'SelectionSet',
 'SelectionDict',
 'SelectionPopup',
 'sendSelectionMenu',
 # ESC menus
 'ESCMenu',
 'sendESCMenu',
 # 'ESCInputMenu',
 # 'sendESCInputMenu',
 'ESCButtonMenu',
 'sendESCButtonMenu',
 'ESCSelectionLanguages',
 'ESCButtonSelectionMenu',
 'sendESCButtonSelectionMenu',
 # Our handy interface
 'CheapPopup',
 ]

###

MAX_DISPLAY_DURATION = 4
DEFAULT_SELECTION_LANGUAGE_PATH = path(es.getAddonPath('_libs')).joinpath('python', 'cheappopup_languages.ini')


lang_ini = cfglib.AddonINI(DEFAULT_SELECTION_LANGUAGE_PATH)

lang_ini.setInitialComments([
 '# ./addons/eventscripts/_libs/python/cheapopup_languages.ini',
 '',
 '# DO NOT translate words with $ or # in front!',
 '# For a list of language abbreviations please see ./addons/eventscripts/_libs/python/deflangs.ini'])

lang_ini.addGroup('header')
lang_ini.setGroupComments('header', 'Default SelectionMenu header')
lang_ini.addValueToGroup('header', 'en', 'Make a selection:')

lang_ini.addGroup('noselections')
lang_ini.setGroupComments('noselections', 'Default SelectionMenu text for no selections')
lang_ini.addValueToGroup('noselections', 'en', 'No selections')

lang_ini.addGroup('separator1')
lang_ini.setGroupComments('separator1', 'Default SelectionMenu separator #1 (separates header from selections)')
lang_ini.addValueToGroup('separator1', 'en', ' ')

lang_ini.addGroup('separator2')
lang_ini.setGroupComments('separator2', 'Default SelectionMenu separator #2 (separates selections from back/next/exit)')
lang_ini.addValueToGroup('separator2', 'en', ' ')

lang_ini.addGroup('back')
lang_ini.setGroupComments('back', 'Default SelectionMenu/ESCButtonSelectionMenu "back" selection text')
lang_ini.addValueToGroup('back', 'en', 'Back')

lang_ini.addGroup('next')
lang_ini.setGroupComments('next', 'Default SelectionMenu/ESCButtonSelectionMenu "next" selection text')
lang_ini.addValueToGroup('next', 'en', 'Next')

lang_ini.addGroup('firstback')
lang_ini.setGroupComments('firstback', 'Default SelectionMenu/ESCButtonSelectionMenu first "back" selection text')
lang_ini.addValueToGroup('firstback', 'en', 'Last page')

lang_ini.addGroup('lastnext')
lang_ini.setGroupComments('lastnext', 'Default SelectionMenu/ESCButtonSelectionMenu last "next" selection text')
lang_ini.addValueToGroup('lastnext', 'en', 'First page')

lang_ini.addGroup('exit')
lang_ini.setGroupComments('exit', 'Default SelectionMenu/ESCButtonSelectionMenu "exit" selection text')
lang_ini.addValueToGroup('exit', 'en', 'Exit')

lang_ini.addGroup('title')
lang_ini.setGroupComments('title', 'Default ESCButtonSelectionMenu title')
lang_ini.addValueToGroup('title', 'en', 'Make a selection:')

lang_ini.addGroup('text')
lang_ini.setGroupComments('text', 'Default ESCButtonSelectionMenu text')
lang_ini.addValueToGroup('text', 'en', 'Please make a selection:')

lang_ini.write()

###

class PopupPlayers(dict):
   def __setitem__(self, userid, value):
      super(PopupPlayers, self).__setitem__(int(userid), value)
      ccfilter.turnOn()

   def __delitem__(self, userid):
      userid = int(userid)
      if userid in self:
         self[userid].cancelDelay()
         super(PopupPlayers, self).__delitem__(userid)
         if not self:
            ccfilter.turnOff()

   def clear(self):
      for userid in self:
         self[userid].cancelDelay()
      super(PopupPlayers, self).clear()
      ccfilter.turnOff()

players = PopupPlayers()

#

class ClientCommandFilter(object):
   active = False

   def turnOn(self):
      if not self.active:
         es.addons.registerClientCommandFilter(self.ccFilter)
      self.active = True

   def turnOff(self):
      if self.active:
         es.addons.unregisterClientCommandFilter(self.ccFilter)
      self.active = False

   def ccFilter(self, userid, args):
      if userid in players and len(args) == 2 and args[0].lower() == 'menuselect':
         players[userid].menuselect(int(args[1]))
      return True

ccfilter = ClientCommandFilter()

###

class Popup(object):
   """
   This is the base Popup class
   All send*Menu functions return an instance or subclass of this class
   """

   def __init__(self, userid, duration, text, callback=None):
      self.userid = int(userid)
      self.duration = int(math.ceil(duration))
      self.text = text
      self.callback = callback
      self.delayname = 'cheappopup_' + str(self.userid)

      if self.userid in players:
         players[self.userid].close()
      players[self.userid] = self
      self.active = False

   def send(self):
      self.active = True
      if not self.duration or self.duration > MAX_DISPLAY_DURATION:
         es.menu(MAX_DISPLAY_DURATION, self.userid, self.text)
         if self.duration:
            self.duration -= MAX_DISPLAY_DURATION
         gamethread.delayedname(MAX_DISPLAY_DURATION, self.delayname, self.send)

      else:
         es.menu(self.duration, self.userid, self.text)
         gamethread.delayedname(self.duration, self.delayname, self.autoClose)

   def autoClose(self):
      self.active = False
      if self.userid in players:
         del players[self.userid]
      if not self.callback is None:
         self.callback(self.userid, None)

   def menuselect(self, item):
      if not self.active:
         return
      self.cancelDelay()
      if not self.callback is None:
         self.callback(self.userid, item)

   def close(self):
      if not self.active:
         return False
      es.menu(1, self.userid, ' ')
      self.cancelDelay() # This cancels any delay and sets "active" to False
      if not self.callback is None:
         try:
            self.callback(self.userid, None)
         finally: pass
      return True

   def cancelDelay(self):
      if not self.active:
         return
      self.active = False
      gamethread.cancelDelayed(self.delayname)

   def isActive(self):
      return self.active

def sendMenu(*a, **kw):
   """
   Sends a single menu to a player
   Optional callback will be called with menu selection
   """
   p = Popup(*a, **kw)
   p.send()
   return p

###

class SelectionLanguages(object):
   header = 'Make a selection:'
   noselections = 'No selections'
   separator1 = ' '
   separator2 = ' '
   back = 'Back'
   next = 'Next'
   firstback = 'Last page'
   lastnext = 'First page'
   exit = 'Exit'

#

class SelectionLanguagesLanglib(object):
   def __init__(self, langobj=None, lang=None, knownas=None):
      self.langobj = langobj
      if langobj is None:
         self.langobj = langlib.Strings(lang_ini)
         self.knownas = None
      else:
         self.knownas = knownas
      self.lang = lang

   def __getattr__(self, item):
      if not self.knownas is None:
         item = self.knownas.get(item, item)
      return self.langobj(item, lang=self.lang)

#

class SelectionSet(set):
   """
   This subclass of "set" remembers the order items are added
   When initialized with an iterator, items are added in sorted(...) order
   """

   def __init__(self, *a, **kw):
      super(SelectionSet, self).__init__(*a, **kw)
      self.order = sorted(self)

   def add(self, item):
      if item in self.order:
         self.order.remove(item)
      self.order.append(item)
      super(SelectionSet, self).add(item)

   def remove(self, item):
      if item in self.order:
         self.order.remove(item)
      super(SelectionSet, self).remove(item)

#

class SelectionDict(dict):
   """
   This subclass of "dict" remembers the order keys are added
   When initialized with an iterator, keys are added in sorted(...) order
   """

   def __init__(self, *a, **kw):
      super(SelectionDict, self).__init__(*a, **kw)
      self.order = sorted(self)

   def __setitem__(self, item, value):
      if item in self.order:
         self.order.remove(item)
      self.order.append(item)

      super(SelectionDict, self).__setitem__(item, value)

   def __delitem__(self, item):
      if item in self.order:
         self.order.remove(item)

      super(SelectionDict, self).__delitem__(item)

#

# We need to subclass Popup to ensure this class is interchangable with Popup in our dictionary
class SelectionPopup(Popup):
   current = None
   page = 0

   def __init__(self, userid, items, callback, langobj=SelectionLanguagesLanglib(),
                      firstback=None, lastnext=None, nohighlight=None):
      self.userid = int(userid)
      self.callback = callback
      self.items = items
      self.langobj = langobj
      self.firstback = firstback
      self.lastnext = lastnext
      self.nohighlight = nohighlight

      # If a string is passed as a language object, use that string as the header
      if isinstance(langobj, basestring):
         self.langobj = SelectionLanguages()
         self.langobj.header = langobj

      self.order = self.items
      # If the object wants to tell us the order (such as SelectionDict or SelectionSet), accept it
      if hasattr(self.items, 'order'):
         self.order = self.items.order
      # If the object isn't ordered itself, sort it
      elif not hasattr(self.items, '__getslice__'): # isinstance(self.items, collections.Sequence):
         self.order = sorted(self.items)
      self.pagecount = max(int(math.ceil(len(self.order) / 7)), 1)

   def autoClose(self):
      # Nobody should ever call this
      pass

   def send(self):
      text = self.langobj.header + '\n' + self.langobj.separator1

      if self.items:
         pageitems = enumerate(self.order[self.page * 7:(self.page + 1) * 7])
         if not self.nohighlight: # Every selection is highlighted (None, False, empty data types)
            for x, i in pageitems:
               text += '\n->' + str(x + 1) + '. ' + str(i)
         elif self.nohighlight is True: # No selection is highlighted
            for x, i in pageitems:
               text += '\n' + str(x + 1) + '. ' + str(i)
         else: # Only given selections go without highlighting
            for x, i in pageitems:
               if i in self.nohighlight:
                  text += '\n' + str(x + 1) + '. ' + str(i)
               else:
                  text += '\n->' + str(x + 1) + '. ' + str(i)
      else:
         text += '\n' + self.langobj.noselections

      text += '\n' + self.langobj.separator2

      if self.page:
         text += '\n8. ' + self.langobj.back
      elif self.pagecount > 1 or not self.firstback is None:
         text += '\n8. ' + self.langobj.firstback

      if self.page == self.pagecount - 1: # Last page
         if self.pagecount > 1 or not self.lastnext is None:
            text += '\n9. ' + self.langobj.lastnext
      else:
         text += '\n9. ' + self.langobj.next

      text += '\n\n0. ' + self.langobj.exit

      self.current = sendMenu(self.userid, 0, text, self.menuselect)

   def menuselect(self, userid, item):
      self.current = None
      if item is None: # Closed menu
         self.callback(userid, None)
         return

      elif item == 8:
         if self.page:
            self.page -= 1
         elif not self.firstback is None:
            self.firstback(userid)
            return
         elif self.pagecount == 1: # Normal missing selection
            self.callback(userid, None)
            return
         else:
            self.page = self.pagecount - 1

      elif item == 9:
         if self.page != self.pagecount - 1:
            self.page += 1
         elif not self.lastnext is None:
            self.lastnext(userid)
            return
         elif self.pagecount == 1: # Normal missing selection
            self.callback(userid, None)
            return
         else:
            self.page = 0

      elif item in (0, 10):
         self.callback(userid, None)
         return

      else:
         if self.page * 7 + item > len(self.order):
            # Selection outside range
            self.callback(userid, None)
         elif hasattr(self.items, '__getslice__'): # isinstance(self.items, collections.Sequence):
            # Selection of ordered items
            self.callback(userid, self.page * 7 + item - 1)
         elif isinstance(self.items, (set, frozenset)): # isinstance(self.items, collections.Set):
            # Selection of unordered items
            self.callback(userid, self.order[self.page * 7 + item - 1])
         else:
            # Selection of mapping
            self.callback(userid, self.items[self.order[self.page * 7 + item - 1]])
         return

      # No selection, so resend menu
      self.send()

   def close(self):
      if self.current is None:
         return False
      r = self.current.close()
      self.current = None
      try:
         self.callback(self.userid, None)
      finally: pass
      return r

   def cancelDelay(self):
      if self.current is None:
         return
      self.current.cancelDelay()
      self.current = None

   def isActive(self):
      if self.current is None:
         return False
      return self.current.isActive()

def sendSelectionMenu(*a, **kw):
   """
   Sends a series of menus to a player prompting selection
   Returns selection through callback
   """
   r = SelectionPopup(*a, **kw)
   r.send()
   return r

###

class ESCMenuPlayers(dict):
   def __init__(self):
      super(ESCMenuPlayers, self).__init__()

      '''
      # If this library or class is reloaded, the most recent load gets the client command
      cmdlib.unregisterClientCommand('_cheappopup_escinput')
      cmdlib.registerClientCommand('_cheappopup_escinput', self.menuselect,
       'Internal CheapPopup client command for catching ESC input text')'''

   def __delitem__(self, userid):
      userid = int(userid)
      if userid in self:
         self[userid].cancelDelay()
         super(ESCMenuPlayers, self).__delitem__(userid)

   def clear(self):
      for userid in self:
         self[userid].close()
      super(ESCMenuPlayers, self).clear()

   '''
   def menuselect(self, userid, args):
      es.msg('menuselect ESCinput: %s, %s' % (userid, args))
      if userid in self:
         self[userid].menuselect(str(args))'''

escplayers = ESCMenuPlayers()

#

class ESCMenu(Popup):
   def __init__(self, userid, duration, title, text):
      self.userid = int(userid)
      self.duration = duration
      self.title = title
      self.text = text
      self.delayname = 'cheappopup_esc_' + str(self.userid)

      if self.userid in escplayers:
         escplayers[self.userid].close()
      escplayers[self.userid] = self
      self.active = False

   def autoClose(self):
      self.active = False

   def send(self):
      self.active = True
      es.esctextbox(self.duration, self.userid, self.title, self.text)
      gamethread.delayedname(self.duration, self.delayname, self.autoClose)

   def menuselect(self, *a, **kw):
      # Nobody should ever call this
      pass

   def close(self):
      if not self.active:
         return False
      es.esctextbox(10, self.userid, self.title, self.text)
      self.cancelDelay()
      return True

   def cancelDelay(self):
      if not self.active:
         return
      self.active = False
      gamethread.cancelDelayed(self.delayname)

   def isActive(self):
      return self.active

def sendESCMenu(*a, **kw):
   """ Sends a menu of text to a player, ESC must be pressed to view """
   r = ESCMenu(*a, **kw)
   r.send()
   return r

###
'''  ESCInputMenu doesn't work as it is no longer allowed to execute commands on clients
sendESCInputMenu(userid, duration, title, text, callback)
   "
   Sends a menu for string input to a player, ESC must be pressed to view
   Returns input through callback
   "
   userid = player to send menu
   duration = integer number of seconds to display menu (10 <= duration <= 200, this limit is Valve-imposed)
   title = title to use for menu (this text also appears to alert the player of the menu)
   text = text to display in menu (prompt for string)
   callback = function to call when a selection is made: callback(userid, selection)
      where "userid" is the integer userid of the player making the selection
      and "selection" is the string the player entered or None for no selection (timeout/menu closed)

class ESCInputMenu(ESCMenu):
   def __init__(self, userid, duration, title, text, callback):
      self.userid = int(userid)
      self.duration = duration
      self.title = title
      self.text = text
      self.callback = callback
      self.delayname = 'cheappopup_esc_' + str(self.userid)

      if self.userid in escplayers:
         escplayers[self.userid].close()
      escplayers[self.userid] = self
      self.active = False

   def autoClose(self):
      self.active = False
      self.callback(self.userid, None)

   def send(self):
      self.active = True
      es.escinputbox(self.duration, self.userid, self.title, self.text, '_cheappopup_escinput')
      gamethread.delayedname(self.duration, self.delayname, self.autoClose)

   def menuselect(self, text):
      self.active = False
      self.cancelDelay()
      self.callback(self.userid, text)

   def close(self):
      r = super(ESCInputMenu, self).close()
      try:
         self.callback(self.userid, None)
      finally: pass
      return r

def sendESCInputMenu(*a, **kw):
   """
   Sends a menu for string input to a player, ESC must be pressed to view
   Returns input through callback
   """
   r = ESCInputMenu(*a, **kw)
   r.send()
   return r
'''
###

class ESCButtonMenu(ESCMenu):
   def __init__(self, userid, duration, title, text, items, callback):
      self.userid = int(userid)
      self.duration = duration
      self.title = title
      self.text = text
      self.items = items
      self.callback = callback
      self.delayname = 'cheappopup_esc_' + str(userid)

      if self.userid in escplayers:
         escplayers[self.userid].close()
      if self.userid in players: # The ESC button menu overlaps the menuselect function of normal popups
         players[self.userid].close()
      players[self.userid] = escplayers[self.userid] = self
      self.active = False

   def autoClose(self):
      self.active = False
      self.callback(self.userid, None)

   def send(self):
      self.active = True
      es.escmenu(self.duration, self.userid, self.title, self.text, *tuple(self.items))
      gamethread.delayedname(self.duration, self.delayname, self.autoClose)

   def menuselect(self, item):
      if not self.active:
         return
      self.cancelDelay()
      self.callback(self.userid, item)

def sendESCButtonMenu(*a, **kw):
   """
   Sends a multi-button GUI menu to the player, ESC must be pressed to view
   Returns selection through callback
   """
   r = ESCButtonMenu(*a, **kw)
   r.send()
   return r

###

class ESCSelectionLanguages(object):
   title = 'Make a selection:'
   text = 'Please make a selection:'
   back = 'Back'
   next = 'Next'
   firstback = 'Last page'
   lastnext = 'First page'

#

class ESCButtonSelectionMenu(ESCMenu):
   current = None
   page = 0

   def __init__(self, userid, duration, items, callback, langobj=SelectionLanguagesLanglib(),
                      firstback=None, lastnext=None):
      self.userid = int(userid)
      self.duration = duration
      self.items = items
      self.callback = callback
      self.langobj = langobj
      self.firstback = firstback
      self.lastnext = lastnext

      # If a string is passed as a language object, use that string as the title and text
      if isinstance(langobj, basestring):
         self.langobj = ESCSelectionLanguages()
         self.langobj.title = self.langobj.text = langobj

      self.order = self.items
      # If the object wants to tell us the order (such as SelectionDict or SelectionSet), accept it
      if hasattr(self.items, 'order'):
         self.order = self.items.order
      # If the object isn't ordered itself, sort it
      elif not hasattr(self.items, '__getslice__'): # isinstance(self.items, collections.Sequence):
         self.order = sorted(self.items)
      self.pagecount = max(int(math.ceil(len(self.order) / 6)), 1)

   def autoClose(self):
      # Nobody should ever call this
      pass

   def send(self):
      buttons = list(self.order[self.page * 6:(self.page + 1) * 6])

      if self.page:
         buttons.append(self.langobj.back)
      elif self.pagecount > 1 or not self.firstback is None:
         buttons.append(self.langobj.firstback)

      if self.page == self.pagecount - 1: # Last page
         if self.pagecount > 1 or not self.lastnext is None:
            buttons.append(self.langobj.lastnext)
      else:
         buttons.append(self.langobj.next)

      self.current = sendESCButtonMenu(self.userid, self.duration,
       self.langobj.title, self.langobj.text, buttons, self.menuselect)

   def menuselect(self, userid, item):
      self.current = None
      backitem = 7
      if self.page == self.pagecount - 1 and len(self.order) % 6:
         backitem = len(self.order) % 6 + 1

      if item is None: # Timeout or closed menu
         self.callback(userid, None)
         return

      elif item == backitem:
         if self.page:
            self.page -= 1
         elif not self.firstback is None:
            self.firstback(userid)
            return
         else:
            self.page = self.pagecount - 1

      elif item == backitem + 1:
         if self.page != self.pagecount - 1:
            self.page += 1
         elif not self.lastnext is None:
            self.lastnext(userid)
            return
         else:
            self.page = 0

      else:
         if self.page * 6 + item > len(self.order):
            # Selection outside range
            self.callback(userid, None)
         elif hasattr(self.items, '__getslice__'): # isinstance(self.items, collections.Sequence):
            # Selection of ordered items
            self.callback(userid, self.page * 6 + item - 1)
         elif isinstance(self.items, (set, frozenset)): # isinstance(self.items, collections.Set):
            # Selection of unordered items
            self.callback(userid, self.order[self.page * 6 + item - 1])
         else:
            # Selection of dictionary-type items
            self.callback(userid, self.items[self.order[self.page * 6 + item - 1]])
         return

      # No selection, so resend menu
      self.send()

   def close(self):
      if self.current is None:
         return False
      r = self.current.close()
      self.current = None
      try:
         self.callback(self.userid, None)
      finally: pass
      return r

   def cancelDelay(self):
      if self.current is None:
         return
      self.current.cancelDelay()
      self.current = None

   def isActive(self):
      if self.current is None:
         return False
      return self.current.isActive()

def sendESCButtonSelectionMenu(*a, **kw):
   """
   Sends a series of multi-button GUI menus to the player prompting selection, ESC must be pressed to view
   Returns selection through callback
   """
   r = ESCButtonSelectionMenu(*a, **kw)
   r.send()
   return r

###

class CheapPopup(object):
   """
   This class is a container for associated menus
   The advantage of this class being the "close" functions only affect menus opened by the same instance
   On player_disconnect any popup delays associated with the disconnecting player are removed
   On es_map_start all popup delays associated with the same instance are removed
   """
   EVENT_NAME_MAP = 'es_map_start'
   EVENT_NAME_DISCONNECT = 'player_disconnect'

   def __init__(self):
      self.players = {}
      self.escplayers = {}

      if self.EVENT_NAME_MAP:
         es.addons.registerForEvent(self, self.EVENT_NAME_MAP, self.mapStart)
      if self.EVENT_NAME_DISCONNECT:
         es.addons.registerForEvent(self, self.EVENT_NAME_DISCONNECT, self.playerDisconnect)

   """ Exposed functions """

   def sendMenu(self, userid, *a, **kw):
      userid = int(userid)
      r = self.players[userid] = sendMenu(userid, *a, **kw)
      return r

   def sendSelectionMenu(self, userid, *a, **kw):
      userid = int(userid)
      r = self.players[userid] = sendSelectionMenu(userid, *a, **kw)
      return r

   def sendESCMenu(self, userid, *a, **kw):
      userid = int(userid)
      r = self.escplayers[userid] = sendESCMenu(userid, *a, **kw)
      return r

   def sendESCInputMenu(self, userid, *a, **kw):
      userid = int(userid)
      r = self.escplayers[userid] = sendESCInputMenu(userid, *a, **kw)
      return r

   def sendESCButtonMenu(self, userid, *a, **kw):
      userid = int(userid)
      # This menu counts as both an ESC menu and a regular popup
      r = self.escplayers[userid] = self.players[userid] = sendESCButtonMenu(userid, *a, **kw)
      return r

   def sendESCButtonSelectionMenu(self, userid, *a, **kw):
      userid = int(userid)
      # This menu counts as both an ESC menu and a regular popup
      r = self.escplayers[userid] = self.players[userid] = sendESCButtonSelectionMenu(userid, *a, **kw)
      return r

   def sendCustomMenu(self, Menu, userid, *a, **kw):
      userid = int(userid)
      r = self.players[userid] = Menu(userid, *a, **kw)
      return r

   def sendCustomESCMenu(self, Menu, userid, *a, **kw):
      userid = int(userid)
      r = self.escplayers[userid] = Menu(userid, *a, **kw)
      return r

   def closeMenu(self, userid, *a, **kw):
      userid = int(userid)
      if userid in self.players:
         return self.players[userid].close(*a, **kw)
      return False

   def closeESCMenu(self, userid, *a, **kw):
      userid = int(userid)
      if userid in self.escplayers:
         return self.escplayers[userid].close(*a, **kw)
      return False

   def closeAll(self, *a, **kw):
      for userid in self.players:
         self.players[userid].close(*a, **kw)
      self.players.clear()

      for userid in self.escplayers:
         self.escplayers[userid].close(*a, **kw)
      self.escplayers.clear()

   def clear(self):
      for userid in self.players:
         self.players[userid].cancelDelay()
      self.players.clear()

      for userid in self.escplayers:
         self.escplayers[userid].cancelDelay()
      self.escplayers.clear()

   """ Event functions """

   def mapStart(self, event_var):
      self.clear()

   def playerDisconnect(self, event_var):
      userid = int(event_var['userid'])
      if userid in self.players:
         self.players[userid].cancelDelay()
         del self.players[userid]

      if userid in self.escplayers:
         self.escplayers[userid].cancelDelay()
         del self.escplayers[userid]

   def unload(self): # This must be called expressly
      self.closeAll()

      if self.EVENT_NAME_MAP:
         es.addons.unregisterForEvent(self, self.EVENT_NAME_MAP)
      if self.EVENT_NAME_DISCONNECT:
         es.addons.unregisterForEvent(self, self.EVENT_NAME_DISCONNECT)

# player_disconnect in L4D is crashy
if es.getGameName() in ('left4dead',):
   CheapPopup.EVENT_NAME_DISCONNECT = None